@@ -10,13 +10,16 @@ use overload
   '""'     => sub { shift->to_string },
   fallback => 1;
 
-use Carp 'croak';
+use Exporter 'import';
 use Mojo::DOM58::_Collection;
 use Mojo::DOM58::_CSS;
-use Mojo::DOM58::_HTML;
+use Mojo::DOM58::_HTML 'tag_to_html';
 use Scalar::Util qw(blessed weaken);
+use Storable 'dclone';
 
-our $VERSION = '1.005';
+our $VERSION = '2.000';
+
+our @EXPORT_OK = 'tag_to_html';
 
 sub new {
   my $class = shift;
@@ -24,11 +27,19 @@ sub new {
   return @_ ? $self->parse(@_) : $self;
 }
 
-sub TO_JSON { shift->_delegate('render') }
+sub new_tag {
+  my $self = shift;
+  my $new  = $self->new;
+  $$new->tag(@_);
+  $$new->xml($$self->xml) if ref $self;
+  return $new;
+}
+
+sub TO_JSON { ${shift()}->render }
 
 sub all_text { _text(_nodes(shift->tree), 1) }
 
-sub ancestors { _select($_[0]->_collect([$_[0]->_ancestors]), $_[1]) }
+sub ancestors { _select($_[0]->_collect([_ancestors($_[0]->tree)]), $_[1]) }
 
 sub append { shift->_add(1, @_) }
 sub append_content { shift->_content(1, 0, @_) }
@@ -78,10 +89,13 @@ sub content {
 
 sub descendant_nodes { $_[0]->_collect(_all(_nodes($_[0]->tree))) }
 
-sub find { $_[0]->_collect($_[0]->_css->select($_[1])) }
+sub find {
+  my $self = shift;
+  return $self->_collect($self->_css->select(@_));
+}
 
-sub following { _select($_[0]->_collect($_[0]->_siblings(1, 1)), $_[1]) }
-sub following_nodes { $_[0]->_collect($_[0]->_siblings(0, 1)) }
+sub following { _select($_[0]->_collect(_siblings($_[0]->tree, 1, 1)), $_[1]) }
+sub following_nodes { $_[0]->_collect(_siblings($_[0]->tree, 0, 1)) }
 
 sub matches { shift->_css->matches(@_) }
 
@@ -92,7 +106,7 @@ sub namespace {
 
   # Extract namespace prefix and search parents
   my $ns = $tree->[1] =~ /^(.*?):/ ? "xmlns:$1" : undef;
-  for my $node ($tree, $self->_ancestors) {
+  for my $node ($tree, _ancestors($tree)) {
 
     # Namespace for prefix
     my $attrs = $node->[2];
@@ -105,8 +119,8 @@ sub namespace {
   return undef;
 }
 
-sub next      { $_[0]->_maybe($_[0]->_siblings(1, 1, 0)) }
-sub next_node { $_[0]->_maybe($_[0]->_siblings(0, 1, 0)) }
+sub next      { $_[0]->_maybe(_siblings($_[0]->tree, 1, 1, 0)) }
+sub next_node { $_[0]->_maybe(_siblings($_[0]->tree, 0, 1, 0)) }
 
 sub parent {
   my $self = shift;
@@ -114,16 +128,16 @@ sub parent {
   return $self->_build(_parent($tree), $self->xml);
 }
 
-sub parse { shift->_delegate(parse => @_) }
+sub parse { ${$_[0]}->parse($_[1]) and return $_[0] }
 
-sub preceding { _select($_[0]->_collect($_[0]->_siblings(1, 0)), $_[1]) }
-sub preceding_nodes { $_[0]->_collect($_[0]->_siblings(0)) }
+sub preceding { _select($_[0]->_collect(_siblings($_[0]->tree, 1, 0)), $_[1]) }
+sub preceding_nodes { $_[0]->_collect(_siblings($_[0]->tree, 0)) }
 
 sub prepend { shift->_add(0, @_) }
 sub prepend_content { shift->_content(0, 0, @_) }
 
-sub previous      { $_[0]->_maybe($_[0]->_siblings(1, 0, -1)) }
-sub previous_node { $_[0]->_maybe($_[0]->_siblings(0, 0, -1)) }
+sub previous      { $_[0]->_maybe(_siblings($_[0]->tree, 1, 0, -1)) }
+sub previous_node { $_[0]->_maybe(_siblings($_[0]->tree, 0, 0, -1)) }
 
 sub remove { shift->replace('') }
 
@@ -135,10 +149,17 @@ sub replace {
 
 sub root {
   my $self = shift;
-  return $self unless my $tree = $self->_ancestors(1);
+  return $self unless my $tree = _ancestors($self->tree, 1);
   return $self->_build($tree, $self->xml);
 }
 
+sub selector {
+  return undef unless (my $tree = shift->tree)->[0] eq 'tag';
+  return join ' > ',
+    reverse map { $_->[1] . ':nth-child(' . (@{_siblings($_, 1)} + 1) . ')' }
+    $tree, _ancestors($tree);
+}
+
 sub strip {
   my $self = shift;
   return $self if (my $tree = $self->tree)->[0] ne 'tag';
@@ -157,9 +178,9 @@ sub tap { Mojo::DOM58::_Collection::tap(@_) }
 
 sub text { _text(_nodes(shift->tree), 0) }
 
-sub to_string { shift->_delegate('render') }
+sub to_string { ${shift()}->render }
 
-sub tree { shift->_delegate(tree => @_) }
+sub tree { @_ > 1 ? (${$_[0]}->tree($_[1]) and return $_[0]) : ${$_[0]}->tree }
 
 sub type { shift->tree->[0] }
 
@@ -189,7 +210,7 @@ sub with_roles { Mojo::DOM58::_Collection::with_roles(@_) }
 sub wrap         { shift->_wrap(0, @_) }
 sub wrap_content { shift->_wrap(1, @_) }
 
-sub xml { shift->_delegate(xml => @_) }
+sub xml { @_ > 1 ? (${$_[0]}->xml($_[1]) and return $_[0]) : ${$_[0]}->xml }
 
 sub _add {
   my ($self, $offset, $new) = @_;
@@ -210,9 +231,9 @@ sub _all {
 }
 
 sub _ancestors {
-  my ($self, $root) = @_;
+  my ($tree, $root) = @_;
 
-  return () unless my $tree = _parent($self->tree);
+  return () unless $tree = _parent($tree);
   my @ancestors;
   do { push @ancestors, $tree }
     while ($tree->[0] eq 'tag') && ($tree = $tree->[3]);
@@ -245,12 +266,7 @@ sub _content {
 
 sub _css { Mojo::DOM58::_CSS->new(tree => shift->tree) }
 
-sub _delegate {
-  my ($self, $method) = (shift, shift);
-  return $$self->$method unless @_;
-  $$self->$method(@_);
-  return $self;
-}
+sub _fragment { _link(my $r = ['root', @_], [@_]); $r }
 
 sub _link {
   my ($parent, $children) = @_;
@@ -282,7 +298,13 @@ sub _offset {
 
 sub _parent { $_[0]->[$_[0][0] eq 'tag' ? 3 : 2] }
 
-sub _parse { Mojo::DOM58::_HTML->new(xml => shift->xml)->parse(shift)->tree }
+sub _parse {
+  my ($self, $input) = @_;
+  return Mojo::DOM58::_HTML->new(xml => $self->xml)->parse($input)->tree
+    unless blessed $input && $input->isa('Mojo::DOM58');
+  my $tree = dclone $input->tree;
+  return $tree->[0] eq 'root' ? $tree : _fragment($tree);
+}
 
 sub _replace {
   my ($self, $parent, $child, $nodes) = @_;
@@ -293,9 +315,9 @@ sub _replace {
 sub _select { $_[1] ? $_[0]->grep(matches => $_[1]) : $_[0] }
 
 sub _siblings {
-  my ($self, $tags, $tail, $i) = @_;
+  my ($tree, $tags, $tail, $i) = @_;
 
-  return defined $i ? undef : [] if (my $tree = $self->tree)->[0] eq 'root';
+  return defined $i ? undef : [] if $tree->[0] eq 'root';
 
   my $nodes = _nodes(_parent($tree));
   my $match = -1;
@@ -385,6 +407,7 @@ Mojo::DOM58 - Minimalistic HTML/XML DOM parser with CSS selectors
 
   # Modify
   $dom->find('div p')->last->append('<p id="c">456</p>');
+  $dom->at('#c')->prepend($dom->new_tag('p', id => 'd', '789'));
   $dom->find(':not(p)')->map('strip');
 
   # Render
@@ -406,7 +429,7 @@ closely compatible with upstream. It differs only in the standalone format and
 compatibility with Perl 5.8. Any bugs or patches not related to these changes
 should be reported directly to the L<Mojolicious> issue tracker.
 
-This release of L<Mojo::DOM58> is up to date with version C<7.62> of
+This release of L<Mojo::DOM58> is up to date with version C<8.09> of
 L<Mojolicious>.
 
 =head1 NODES AND ELEMENTS
@@ -538,6 +561,13 @@ An C<E> element whose C<foo> attribute value contains the substring C<bar>.
   my $contains = $dom->find('input[name*="fo"]');
   my $contains = $dom->find('input[name*=fo]');
 
+=item E[foo|="en"]
+
+An C<E> element whose C<foo> attribute has a hyphen-separated list of values
+beginning (from the left) with C<en>.
+
+  my $english = $dom->find('link[hreflang|=en]');
+
 =item E:root
 
 An C<E> element, root of the document.
@@ -622,6 +652,20 @@ An C<E> element that has no children (including text nodes).
 
   my $empty = $dom->find(':empty');
 
+=item E:link
+
+An C<E> element being the source anchor of a hyperlink of which the target is
+not yet visited (C<:link>) or already visited (C<:visited>). Note that
+L<Mojo::DOM58> is not stateful, therefore C<:link> and C<:visited> yield
+exactly the same results.
+
+  my $links = $dom->find(':link');
+  my $links = $dom->find(':visited');
+
+=item E:visited
+
+Alias for L</"E:link">.
+
 =item E:checked
 
 A user interface element C<E> which is checked (for instance a radio-button or
@@ -664,6 +708,19 @@ This selector is part of
 L<Selectors Level 4|http://dev.w3.org/csswg/selectors-4>, which is still a work
 in progress.
 
+=item A|E
+
+An C<E> element that belongs to the namespace alias C<A> from
+L<CSS Namespaces Module Level 3|https://www.w3.org/TR/css-namespaces-3/>.
+Key/value pairs passed to selector methods are used to declare namespace
+aliases.
+
+  my $elem = $dom->find('lq|elem', lq => 'http://example.com/q-markup');
+
+Using an empty alias searches for an element that belongs to no namespace.
+
+  my $div = $dom->find('|div');
+
 =item E F
 
 An C<F> element descendant of an C<E> element.
@@ -736,6 +793,19 @@ Alias for L</"attr">.
 
 Alias for L</"to_string">.
 
+=head1 FUNCTIONS
+
+L<Mojo::DOM58> implements the following functions, which can be imported
+individually.
+
+=head2 tag_to_html
+
+  my $str = tag_to_html 'div', id => 'foo', 'safe content';
+
+Generate HTML/XML tag and render it right away. This is a significantly faster
+alternative to L</"new_tag"> for template systems that have to generate a lot
+of tags.
+
 =head1 METHODS
 
 L<Mojo::DOM58> implements the following methods.
@@ -748,6 +818,45 @@ L<Mojo::DOM58> implements the following methods.
 Construct a new scalar-based L<Mojo::DOM58> object and L</"parse"> HTML/XML
 fragment if necessary.
 
+=head2 new_tag
+
+  my $tag = Mojo::DOM58->new_tag('div');
+  my $tag = $dom->new_tag('div');
+  my $tag = $dom->new_tag('div', id => 'foo', hidden => undef);
+  my $tag = $dom->new_tag('div', 'safe content');
+  my $tag = $dom->new_tag('div', id => 'foo', 'safe content');
+  my $tag = $dom->new_tag('div', data => {mojo => 'rocks'}, 'safe content');
+  my $tag = $dom->new_tag('div', id => 'foo', sub { 'unsafe content' });
+
+Construct a new L<Mojo::DOM58> object for an HTML/XML tag with or without
+attributes and content. The C<data> attribute may contain a hash reference with
+key/value pairs to generate attributes from.
+
+  # "<br>"
+  $dom->new_tag('br');
+
+  # "<div></div>"
+  $dom->new_tag('div');
+
+  # "<div id="foo" hidden></div>"
+  $dom->new_tag('div', id => 'foo', hidden => undef);
+
+  # "<div>test &amp; 123</div>"
+  $dom->new_tag('div', 'test & 123');
+
+  # "<div id="foo">test &amp; 123</div>"
+  $dom->new_tag('div', id => 'foo', 'test & 123');
+
+  # "<div data-foo="1" data-bar="test">test &amp; 123</div>""
+  $dom->new_tag('div', data => {foo => 1, Bar => 'test'}, 'test & 123');
+
+  # "<div id="foo">test & 123</div>"
+  $dom->new_tag('div', id => 'foo', sub { 'test & 123' });
+
+  # "<div>Hello<b>Mojo!</b></div>"
+  $dom->parse('<div>Hello</div>')->at('div')
+    ->append_content($dom->new_tag('b', 'Mojo!'))->root;
+
 =head2 all_text
 
   my $text = $dom->all_text;
@@ -772,6 +881,7 @@ objects. All selectors listed in L</"SELECTORS"> are supported.
 =head2 append
 
   $dom = $dom->append('<p>I ♥ Mojo::DOM58!</p>');
+  $dom = $dom->append(Mojo::DOM58->new);
 
 Append HTML/XML fragment to this node (for all node types other than C<root>).
 
@@ -786,6 +896,7 @@ Append HTML/XML fragment to this node (for all node types other than C<root>).
 =head2 append_content
 
   $dom = $dom->append_content('<p>I ♥ Mojo::DOM58!</p>');
+  $dom = $dom->append_content(Mojo::DOM58->new);
 
 Append HTML/XML fragment (for C<root> and C<tag> nodes) or raw content to this
 node's content.
@@ -804,6 +915,7 @@ node's content.
 =head2 at
 
   my $result = $dom->at('div ~ p');
+  my $result = $dom->at('svg|line', svg => 'http://www.w3.org/2000/svg');
 
 Find first descendant element of this element matching the CSS selector and
 return it as a L<Mojo::DOM58> object, or C<undef> if none could be found. All
@@ -812,6 +924,12 @@ selectors listed in L</"SELECTORS"> are supported.
   # Find first element with "svg" namespace definition
   my $namespace = $dom->at('[xmlns\:svg]')->{'xmlns:svg'};
 
+Trailing key/value pairs can be used to declare xml namespace aliases.
+
+  # "<rect />"
+  $dom->parse('<svg xmlns="http://www.w3.org/2000/svg"><rect /></svg>')
+    ->at('svg|rect', svg => 'http://www.w3.org/2000/svg');
+
 =head2 attr
 
   my $hash = $dom->attr;
@@ -862,6 +980,7 @@ objects. All selectors listed in L</"SELECTORS"> are supported.
 
   my $str = $dom->content;
   $dom    = $dom->content('<p>I ♥ Mojo::DOM58!</p>');
+  $dom    = $dom->content(Mojo::DOM58->new);
 
 Return this node's content or replace it with HTML/XML fragment (for C<root>
 and C<tag> nodes) or raw content.
@@ -905,6 +1024,7 @@ this element as L<Mojo::DOM58> objects.
 =head2 find
 
   my $collection = $dom->find('div ~ p');
+  my $collection = $dom->find('svg|line', svg => 'http://www.w3.org/2000/svg');
 
 Find all descendant elements of this element matching the CSS selector and
 return a L<collection|/"COLLECTION METHODS"> containing these elements as
@@ -922,6 +1042,12 @@ L<Mojo::DOM58> objects. All selectors listed in L</"SELECTORS"> are supported.
   # Find elements with a class that contains dots
   my @divs = $dom->find('div.foo\.bar')->each;
 
+Trailing key/value pairs can be used to declare xml namespace aliases.
+
+  # "<rect />"
+  $dom->parse('<svg xmlns="http://www.w3.org/2000/svg"><rect /></svg>')
+    ->find('svg|rect', svg => 'http://www.w3.org/2000/svg')->first;
+
 =head2 following
 
   my $collection = $dom->following;
@@ -947,6 +1073,7 @@ this node as L<Mojo::DOM58> objects.
 =head2 matches
 
   my $bool = $dom->matches('div ~ p');
+  my $bool = $dom->matches('svg|line', svg => 'http://www.w3.org/2000/svg');
 
 Check if this element matches the CSS selector. All selectors listed in
 L</"SELECTORS"> are supported.
@@ -959,6 +1086,12 @@ L</"SELECTORS"> are supported.
   $dom->parse('<p class="a">A</p>')->at('p')->matches('.b');
   $dom->parse('<p class="a">A</p>')->at('p')->matches('p[id]');
 
+Trailing key/value pairs can be used to declare xml namespace aliases.
+
+  # True
+  $dom->parse('<svg xmlns="http://www.w3.org/2000/svg"><rect /></svg>')
+    ->matches('svg|rect', svg => 'http://www.w3.org/2000/svg');
+
 =head2 namespace
 
   my $namespace = $dom->namespace;
@@ -1040,6 +1173,7 @@ before this node as L<Mojo::DOM58> objects.
 =head2 prepend
 
   $dom = $dom->prepend('<p>I ♥ Mojo::DOM58!</p>');
+  $dom = $dom->prepend(Mojo::DOM58->new);
 
 Prepend HTML/XML fragment to this node (for all node types other than C<root>).
 
@@ -1054,6 +1188,7 @@ Prepend HTML/XML fragment to this node (for all node types other than C<root>).
 =head2 prepend_content
 
   $dom = $dom->prepend_content('<p>I ♥ Mojo::DOM58!</p>');
+  $dom = $dom->prepend_content(Mojo::DOM58->new);
 
 Prepend HTML/XML fragment (for C<root> and C<tag> nodes) or raw content to this
 node's content.
@@ -1110,6 +1245,7 @@ Remove this node and return L</"root"> (for C<root> nodes) or L</"parent">.
 =head2 replace
 
   my $parent = $dom->replace('<div>I ♥ Mojo::DOM58!</div>');
+  my $parent = $dom->replace(Mojo::DOM58->new);
 
 Replace this node with HTML/XML fragment and return L</"root"> (for C<root>
 nodes) or L</"parent">.
@@ -1127,6 +1263,18 @@ nodes) or L</"parent">.
 
 Return L<Mojo::DOM58> object for C<root> node.
 
+=head2 selector
+
+  my $selector = $dom->selector;
+
+Get a unique CSS selector for this element.
+
+  # "ul:nth-child(1) > li:nth-child(2)"
+  $dom->parse('<ul><li>Test</li><li>123</li></ul>')->find('li')->last->selector;
+
+  # "p:nth-child(1) > b:nth-child(1) > i:nth-child(1)"
+  $dom->parse('<p><b><i>Test</i></b></p>')->at('i')->selector;
+
 =head2 strip
 
   my $parent = $dom->strip;
@@ -1254,6 +1402,7 @@ L<Role::Tiny> (2.000001+).
 =head2 wrap
 
   $dom = $dom->wrap('<div></div>');
+  $dom = $dom->wrap(Mojo::DOM58->new);
 
 Wrap HTML/XML fragment around this node (for all node types other than C<root>),
 placing it as the last child of the first innermost element.
@@ -1273,6 +1422,7 @@ placing it as the last child of the first innermost element.
 =head2 wrap_content
 
   $dom = $dom->wrap_content('<div></div>');
+  $dom = $dom->wrap_content(Mojo::DOM58->new);
 
 Wrap HTML/XML fragment around this node's content (for C<root> and C<tag>
 nodes), placing it as the last children of the first innermost element.
@@ -1493,7 +1643,8 @@ Turn collection into array reference.
 
 Create a new L<collection|/"COLLECTION METHODS"> without duplicate elements,
 using the string representation of either the elements or the return value of
-the callback/method.
+the callback/method to decide uniqueness. Note that C<undef> and empty string
+are treated the same.
 
   # Longer version
   my $new = $collection->uniq(sub { $_->$method(@args) });
